import datetime
import logging
import time

import dateutil.parser

from ..circuit_serialization import encode_circuit
from ..exceptions import DeqartBaseException

logger = logging.getLogger("deqart-python-sdk")

_SINGLE_REQUEST_LIMIT = 100


def search_jobs(_connection, run_status=None, created_later_than=None):
    if created_later_than is None:
        parsed_created_later_than = None
    elif isinstance(created_later_than, str):
        parsed_created_later_than = dateutil.parser.parse(created_later_than)
        if parsed_created_later_than.tzinfo is None:
            logger.warning(
                "created_later_than is a str object without timezone info, assuming UTC timezone"
            )
            parsed_created_later_than = parsed_created_later_than.replace(
                tzinfo=datetime.timezone.utc
            )
        parsed_created_later_than = parsed_created_later_than.isoformat()
    elif isinstance(created_later_than, datetime.datetime):
        if created_later_than.tzinfo is None:
            raise DeqartBaseException(
                0, "created_later_than is a datetime object without timezone info"
            )
        parsed_created_later_than = created_later_than.isoformat()
    else:
        raise DeqartBaseException(
            0, "created_later_than should be None, str, or datetime.datetime object"
        )

    params = {
        "limit": _SINGLE_REQUEST_LIMIT,
        "run_status": run_status,
        "created_later_than": parsed_created_later_than,
    }
    result_dict = {"data": []}
    while True:
        response = _connection.send_request(req_type="GET", path="/jobs", params=params)
        if not response.ok:
            raise DeqartBaseException(
                response.status_code, "Couldn't search jobs " + response.text
            )
        response = response.json()

        results = response["data"]
        result_dict["data"] += results
        result_dict["total_count"] = response["total_count"]

        if len(results) < _SINGLE_REQUEST_LIMIT:
            break

        params["offset"] = len(result_dict["data"])
    return result_dict


def estimate_job_runtime(_connection, circuit):
    encoded_circuit = encode_circuit(circuit)
    params = {"estimate_only": True, "circuit": encoded_circuit}
    response = _connection.send_request(req_type="POST", path="/jobs", json_req=params)
    if not response.ok:
        raise DeqartBaseException(
            response.status_code, "Couldn't estimate job " + response.text
        )
    return response.json()


def submit_job(_connection, circuit, job_name=None, debug_size=None):
    encoded_circuit = encode_circuit(circuit)
    params = {"circuit": encoded_circuit, "job_name": job_name}
    if debug_size is not None:
        params["debug_size"] = debug_size
    response = _connection.send_request(req_type="POST", path="/jobs", json_req=params)
    if not response.ok:
        raise DeqartBaseException(
            response.status_code, "Couldn't submit job " + response.text
        )
    return response.json()["data"]


def cancel_job(_connection, job_id):
    response = _connection.send_request(
        req_type="PATCH", path=f"/jobs/{job_id}", json_req={"cancel": True}
    )
    if not response.ok:
        raise DeqartBaseException(
            response.status_code, "Couldn't cancel job " + response.text
        )
    return response.json()["data"]


def wait_for_job(_connection, job_id):
    i = 1
    while True:
        response = _connection.send_request(req_type="GET", path=f"/jobs/{job_id}")
        if not response.ok:
            raise DeqartBaseException(
                response.status_code, "Couldn't search jobs " + response.text
            )
        response = response.json()["data"]
        if response["run_status"] not in [
            "COMPLETED",
            "CANCELED",
            "TERMINATED",
            "NOT_ENOUGH_FUNDS",
        ]:
            if i < 100:
                time.sleep(0.5)
            else:
                time.sleep(10.0)
        else:
            return response
        i += 1



class EstimateResult:
    """This class contains information about the estimated runtime/cost that
    will be incurred from running the given circuit on the given quantum
    machine/simulator. WARNING: this is just an estimate, the actual runtime
    may be less or more."""

    def __init__(self, data):
        self.device = data.get("device")
        self.estimate_ms = data.get("estimate_ms")
        self.num_qubits = data.get("num_qubits")
        self.circuit = data.get("qc")
        self.warning_message = data.get("warning_message")

    def __str__(self):
        return f"Estimation time in milliseconds: {self.estimate_ms}"



import collections
import datetime
from io import BytesIO

import numpy as np
import requests
from dateutil.parser import parse


class JobResult:
    """This class contains information from a job run on quantum hardware or
    quantum simulators. Mainly it contains the resulting statevector from the
    run. It might contain only partial information, such as job name,
    when dq.run() is called with `asynchronous=true`."""

    def __init__(self, data):
        #  "PENDING" | "QUEUED" |"RUNNING" | "TERMINATED" | "CANCELED" | "OUT_OF_FUNDS" | "COMPLETED"
        self.run_status = data.get("run_status")
        self.job_id = data.get("job_id")
        self.job_name = data.get("job_name")
        self.sv_link = data.get("sv_link")
        self.top_100_results = data.get("top_100_results")
        self.num_qubits = data.get("num_qubits")
        self.circuit = data.get("qc")
        self.tags = data.get("tags")

        self.created_on = data.get("created_on")

        self.queue_start = data.get("queue_start")
        if self.queue_start is not None:
            self.queue_start = parse(self.queue_start)

        self.run_start = data.get("run_start")
        if self.run_start is not None:
            self.run_start = parse(self.run_start)

        self.run_end = data.get("run_end")
        if self.run_end is not None:
            self.run_end = parse(self.run_end)

        if self.queue_start is not None and self.run_start is not None:
            self.queue_time_ms = int(
                (self.run_start - self.queue_start) / datetime.timedelta(milliseconds=1)
            )
        else:
            self.queue_time_ms = None

        if self.run_start is not None and self.run_end is not None:
            self.run_time_ms = round(
                (self.run_end - self.run_start) / datetime.timedelta(milliseconds=1)
            )
        else:
            self.run_time_ms = None

        self.error_message = data.get("error_message")

        self.cost = data.get("cost")
        if self.cost is not None:
            self.cost /= 100.0

    def get_statevector(self):
        response = requests.get(self.sv_link + "/statevector.txt", timeout=100)
        if response.ok:
            data = response.content
            result = np.loadtxt(BytesIO(data), dtype=np.complex_)
            return result
        return None

    def get_counts(self):
        # TODO it's misleading to call this sv_link. It should have been result_link
        if self.sv_link is None:
            return None
        response = requests.get(self.sv_link + "/metadata.json", timeout=100)
        data = response.json()
        if "counts" not in data:
            raise Exception("The job result metadata doesn't contain counts.")
        return data["counts"]

    def __str__(self):
        return f"Job ID: {self.job_id}, name: {self.job_name}, run status: {self.run_status}, queue time (ms): {self.queue_time_ms}, run time (ms): {self.run_time_ms}, cost ($): {self.cost}, num qubits: {self.num_qubits}, error_message: {self.error_message}"
